
#include "quant.h"

/*****************************************************************
  Function to read data in from a file written in free format and to
  form a data matrix for the pedigree.
  This version uses the "genetic data columns" as fixed-effects indicators,
  and also reads in a quantitative data value "quant" -- becomes "x"
  The nghds() and sortdat() routines are then called

  Various checks on pedigree size etc. are carried out. The global variables
  "pedsize" and "namestring" are also evaluated here.

   This new version installed in the Quant library; 8/31/91
   Modified for greater flexibility; 9/29/91 -- same routines now
                used by univar, bivar, and also by mixd

   Modified to use command-line specified data-file;  4/22/92.
   Modified 12/92 (SL) to read additional variables
   Modified 2/93 (EAT) to make compatible with other data inputs, again.
   Modified 6/93 (EAT) to transfer pedigree input to Nghds.
   Modified 8/96 (MAJ) for improved output data format

*****************************************************************/

static void sortdat (double **, int);
static void sortfixed (int);
static int missfixed (int);
static void printfixed (int);

int   input_data (FILE * fp, int nn, int **dodo)
{
   int   j;
   double **quant;

   if (NTraits > 0)
   {
      quant = malloc (NTraits * sizeof (double *));
      alloc_chk (WHERE, quant);
      quant[0] = malloc (PedSize * NTraits * sizeof (double));
      alloc_chk (WHERE, quant[0]);
      for (j = 1; j < NTraits; ++j)
         quant[j] = quant[j - 1] + PedSize;
   }
   else
   {
      quant = NULL;
   }
   input_ped (fp, nn, dodo, quant);

   /* Now allocate effects etc. , since pedsize now known          */
   /* ("PedSize" is maximum pedigree size, "pedsize" is size read) */

   if (NTraits > 0)
   {
      x = malloc (NTraits * sizeof (double *));
      alloc_chk (WHERE, x);
      for (j = 0; j < NTraits; ++j)
      {
         x[j] = malloc ((pedsize + 1) * sizeof (double));
         alloc_chk (WHERE, x[j]);
      }

      /* This often seems to be getting unassigned - SCH */
      for (j = 0; j < NTraits; j++)
         x[j][0] = 0.0;

      qdex = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, qdex);
      idex = malloc ((pedsize + 1) * sizeof (int));
      alloc_chk (WHERE, idex);
      for (j = 0; j <= NFix; ++j)
      {
         fixdex[j] = malloc ((pedsize + 1) * sizeof (int));
         alloc_chk (WHERE, fixdex[j]);
      }

      /* Now sort the fixed effects  */
      sortdat (quant, NTraits);
      printfixed (NTraits);

      /*
       * Now done with quant ; observations are in x; and idex will permit
       * restoring original indexing if necessary
       */
      free ((void *) quant[0]);
      quant[0] = NULL;
      free ((void *) quant);
      quant = NULL;

      /* x, idex and fixdex are remalloced for observations  */
      for (j = 0; j < NTraits; ++j)
      {
         x[j] = realloc (x[j], (np + 1) * sizeof (double));
         alloc_chk (WHERE, x[j]);
      }
      idex = realloc (idex, (np + 1) * sizeof (int));
      alloc_chk (WHERE, idex);

      for (j = 0; j <= NFix; ++j)
      {
         fixdex[j] = realloc (fixdex[j], (np + 1) * sizeof (int));
         alloc_chk (WHERE, fixdex[j]);
      }
   }
   return (NTraits);
}

/*****************************************************************/
/* Sorting subroutines for data and effects                      */
/*****************************************************************/

static void sortdat (double **qq, int nt)
{
   int   i, j, k, qi;

   for (i = 0; i <= NFix; ++i)
      IFix[i] = 1;
   for (i = 0; i < pedsize; ++i)
   {
      qi = 0;
      for (k = 0; k < nt; ++k)
         if ((int) qq[k][i] == MISSING)
            qi = MISSING;

/* Modified EAT July 2002: first check whether covariate info is missing */
      if (missfixed (i))
         qi = MISSING;

      if (qi != MISSING)
      {
         ++np;
         idex[np] = i;
         qdex[i] = np;
         for (j = 0; j < nt; ++j)
            x[j][np] = qq[j][i];
         sortfixed (i);
      }
      else
         qdex[i] = 0;
   }
   return;
}

static int missfixed (int i)
/* routine added by EAT July 2002, to check for missing covariate info */
{
   int   mm, j, k;
   mm = 0;

   if (FESex)
      k = 3;
   else
      k = 4;

   for (j = 1; j <= NFix; ++j)
   {
      if (data[i][j + k] == 0)
         mm = 1;
   }
   return (mm);
}

static void sortfixed (int i)
{                                       /* routine to set fixed effects for
                                         * np'th data individual; ith by name */
   int   j, k;

   if (FESex)
      k = 3;
   else
      k = 4;

   fixdex[0][np] = 1;                   /* overall mean effect */
   for (j = 1; j <= NFix; ++j)
   {
      fixdex[j][np] = data[i][j + k];
      if (fixdex[j][np] < 0)
         give_up (ERQ_nfe, FENo[j], nghd[i].self);
      else if (fixdex[j][np] > LEVELS)
         give_up (ERQ_rfe, LEVELS, FENo[j], nghd[i].self);
   }
   return;
}

static void printfixed (int nt)
{
   int   i, j, k;
   int   nf[LEVELS + 1];

   printf ("% d individuals are observed for", np);
   if (nt == 1)
      printf (" 1 quantitative trait");
   else if (nt > 1)
      printf (" %d quantitative traits", nt);
   if (NFix == 1)
      printf (" and 1 fixed effect");
   else if (NFix > 1)
      printf (" and %d fixed effects", NFix);
   printf ("\n\n");
   fflush (stdout);

   for (i = 1; i <= NFix; ++i)
      IFix[i] = intmax (fixdex[i], 1, np);

   if (NFix > 0)
   {
      printf (" fixed    number       number observed                  ");
      printf ("present in trait\n");
      printf (" effect   of levels    in each level                    ");
      for (j = 0; j < nt; ++j)
         printf ("%d   ", TrNo[j]);
      printf ("\n ------   ---------    ---------------                  ");
      printf ("----------------");
      for (j = 1; j <= NFix; ++j)
      {
         for (i = 1; i <= IFix[j]; ++i)
            nf[i] = 0;
         for (i = 1; i <= np; ++i)
            ++nf[fixdex[j][i]];
         printf ("\n %4d     %4d        ", FENo[j], IFix[j]);
         for (i = 1; i <= IFix[j]; ++i)
         {
            if (i != 1 && i - (i / 6) * 6 == 1)
               printf ("\n                      ");
            printf ("%5d", nf[i]);
            if ((i <= 6 && i == IFix[j]) || i == 6)
            {
               for (k = 1; k <= 6 - i; ++k)
                  printf ("     ");
               printf (" ");
               for (k = 0; k < nt; ++k)
                  if (FfDex[k][j] == 0)
                     printf ("   N");
                  else
                     printf ("   Y");
            }
         }
      }
      printf ("\n\n");
      fflush (stdout);
   }
   return;
}
